/*
 *  Copyright (C) 2006-10 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "nprobe.h"

#define BASE_ID            NTOP_BASE_ID+180
#define URL_MAX_LEN        64
#define HOST_MAX_LEN       32

#define MAX_BYTES_SENT     64
#define MAX_BYTES_RCVD     64

/* RFC 2616 */
/* Request */
#define GET_URL            "GET /"
#define POST_URL           "POST /"
#define HEAD_URL           "HEAD /"
/* Note the other methods (PUT, DELETE, TRACE, CONNECT)
   have not been implemented as they are used very seldom */
#define HOST_STR           "Host: "

/* Response */
#define HTTP_1_0_URL       "HTTP/1.0 "
#define HTTP_1_1_URL       "HTTP/1.1 "

static V9V10TemplateElementId httpPlugin_template[] = {
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID,   VARIABLE_FIELD_LEN, URL_MAX_LEN, ascii_format, dump_as_ascii, "HTTP_URL", "HTTP URL" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+1, STATIC_FIELD_LEN, 2, numeric_format, dump_as_uint, "HTTP_RET_CODE", "HTTP return code (e.g. 200, 304...)" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, 0, STATIC_FIELD_LEN, 0, 0, 0, NULL, NULL }
};

struct plugin_info {
  char *http_url, *http_host;
  uint16_t ret_code;
};

// #define DEBUG 

/* *********************************************** */

static PluginInfo httpPlugin; /* Forward */

/* ******************************************* */

void httpPlugin_init(int argc, char *argv[]) {
  traceEvent(TRACE_INFO, "Initialized HTTP plugin");
}

/* *********************************************** */

/* Handler called whenever an incoming packet is received */

static void httpPlugin_packet(u_char new_bucket, void *pluginData,
			      FlowHashBucket* bkt,
			      u_short proto, u_char isFragment,
			      u_short numPkts, u_char tos,
			      u_short vlanId, struct ether_header *ehdr,
			      IpAddress *src, u_short sport,
			      IpAddress *dst, u_short dport,
			      uint len, uint8_t flags, uint8_t icmpType,
			      u_short numMplsLabels,
			      u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN],
			      char *fingerprint,
			      const struct pcap_pkthdr *h, const u_char *p,
			      u_char *payload, int payloadLen) {
  PluginInformation *info;
  struct plugin_info *pinfo;

  // traceEvent(TRACE_INFO, "httpPlugin_packet(%d)", payloadLen);

  if(new_bucket) {
    info = (PluginInformation*)malloc(sizeof(PluginInformation));
    if(info == NULL) {
      	traceEvent(TRACE_ERROR, "Not enough memory?");
	return; /* Not enough memory */
    }

    info->pluginPtr  = (void*)&httpPlugin;
    pluginData = info->pluginData = malloc(sizeof(struct plugin_info));

    if(info->pluginData == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      free(info);
      return; /* Not enough memory */
    } else
      memset(info->pluginData, 0, sizeof(struct plugin_info));

    info->next = bkt->plugin;
    bkt->plugin = info;
  }

  pinfo = (struct plugin_info*)pluginData;

  if(payloadLen > 0) {
    char *method;
    
    //traceEvent(TRACE_INFO, "==> [%d][%d]'%s'", bkt->bytesSent, bkt->bytesRcvd, payload);

    if((!strncmp((char*)payload, GET_URL, strlen(GET_URL)))) method = GET_URL;
    else if((!strncmp((char*)payload, POST_URL, strlen(POST_URL)))) method = POST_URL;
    else if((!strncmp((char*)payload, HTTP_1_0_URL, strlen(HTTP_1_0_URL)))) method = HTTP_1_0_URL;
    else if((!strncmp((char*)payload, HTTP_1_1_URL, strlen(HTTP_1_1_URL)))) method = HTTP_1_1_URL;
    else method = NULL;

    if(method) {
      char url[1024], *host;
      int i, displ;
      
      if((!strncmp(method, GET_URL, strlen(GET_URL))) 
	 || (!strncmp(method, POST_URL, strlen(POST_URL)))) {
	/* We need to export this flow now */
	if(pinfo->http_url != NULL) {
	  exportBucket(bkt, 0);
	  resetBucketStats(bkt, h, len, sport, dport, payload, payloadLen);
	  free(pinfo->http_url);
	  pinfo->http_url = NULL;
	}

	displ = 1;
      } else
	displ = 0;

      strncpy(url, (char*)&payload[strlen(method)-displ], 
	      min(sizeof(url), (payloadLen-(strlen(method)-displ))));

      url[sizeof(url)-1] = '\0';
      for(i=0; i<sizeof(url); i++) 
	if((url[i] == ' ') 
	   || (url[i] == '\r')
	   || (url[i] == '\n')) {
	  url[i] = '\0';
	  break;
	}

      host = strstr((char*)payload, HOST_STR);
      if(host != NULL) {
	if(pinfo->http_host != NULL) free(pinfo->http_host);

	pinfo->http_host = strdup(&host[sizeof(HOST_STR)-1]);

	if(pinfo->http_host) {
	  for(i=0; pinfo->http_host[i] != '\0'; i++)
	    if((pinfo->http_host[i] == ' ')
	       || (pinfo->http_host[i] == '\r')
	       || (pinfo->http_host[i] == '\n')) {
	      pinfo->http_host[i] = '\0';
	      break;
	    }

#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> HOST='%s'", pinfo->http_host);
#endif
	}
      }

      if(displ == 1) {
	if(pinfo->http_url) free(pinfo->http_url);
	pinfo->http_url = strdup(url);
    } else
	pinfo->ret_code = atoi(url);
    }
  }
}

/* *********************************************** */

/* Handler called when the flow is deleted (after export) */

static void httpPlugin_delete(FlowHashBucket* bkt, void *pluginData) {
  if(pluginData != NULL) {
    struct plugin_info *pinfo = (struct plugin_info*)pluginData;
    
    if(pinfo->http_url)  free(pinfo->http_url);
    if(pinfo->http_host) free(pinfo->http_host);

    free(pluginData);
  }
}

/* *********************************************** */
   
/* Handler called at startup when the template is read */

static V9V10TemplateElementId* httpPlugin_get_template(char* template_name) {
  int i;

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(!strcmp(template_name, httpPlugin_template[i].templateElementName)) {
      return(&httpPlugin_template[i]);
    }
  }

  return(NULL); /* Unknown */
}

/* *********************************************** */

static int get_url(struct plugin_info *info, char *url_buf, 
		   uint urlbuf_len, uint16_t compressData) {  
  return(snprintf(url_buf, urlbuf_len, "%s%s",
		  info->http_host ? info->http_host : "",
		  info->http_url ? info->http_url : "" ));
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be exported */

static int httpPlugin_export(void *pluginData, V9V10TemplateElementId *theTemplate,
			     int direction /* 0 = src->dst, 1 = dst->src */,
			     FlowHashBucket *bkt, char *outBuffer,
			     uint* outBufferBegin, uint* outBufferMax) {
  int i;

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == httpPlugin_template[i].templateElementId) {
      if((*outBufferBegin)+httpPlugin_template[i].templateElementLen > (*outBufferMax))
	return(-2); /* Too long */

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	char url_buf[512] = { 0 };
	uint len;

	switch(httpPlugin_template[i].templateElementId) {
	case BASE_ID:
	  get_url(info, url_buf, sizeof(url_buf)-1, 1);

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (httpPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(url_buf), httpPlugin_template[i].templateElementLen);
	    
	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }    
	  } else
	    len = httpPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], url_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> URL='%s'", url_buf);
#endif
	  (*outBufferBegin) += len;
	  break;
	case BASE_ID+1:
	  copyInt16(info->ret_code, outBuffer, outBufferBegin, outBufferMax);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> RetCode='%d'", info->ret_code);
#endif
	  break;
	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be printed on file */

static int httpPlugin_print(void *pluginData, V9V10TemplateElementId *theTemplate,
			    int direction /* 0 = src->dst, 1 = dst->src */,
			    FlowHashBucket *bkt, char *line_buffer, uint line_buffer_len) {
  int i;
  char url_buf[256] = { 0 };

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == httpPlugin_template[i].templateElementId) {

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	
	switch(httpPlugin_template[i].templateElementId) {
	case BASE_ID:
	  get_url(info, url_buf, sizeof(url_buf), 0);
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", url_buf);
	  break;
	case BASE_ID+1:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->ret_code);
	  break;
	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

static V9V10TemplateElementId* httpPlugin_conf(void) {
  return(httpPlugin_template);
}

/* *********************************************** */

/* Plugin entrypoint */
static PluginInfo httpPlugin = {
  NPROBE_REVISION,
  "HTTP Protocol Dissector",
  "0.1",
  "Handle HTTP protocol",
  "L.Deri <deri@ntop.org>",
  0 /* not always enabled */, 1, /* enabled */
  httpPlugin_init,
  NULL, /* Term */
  httpPlugin_conf,
  httpPlugin_delete,
  1, /* call packetFlowFctn for each packet */
  httpPlugin_packet,
  httpPlugin_get_template,
  httpPlugin_export,
  httpPlugin_print,
  NULL,
  NULL
};

/* *********************************************** */

/* Plugin entry fctn */
#ifdef MAKE_STATIC_PLUGINS
PluginInfo* httpPluginEntryFctn(void)
#else
  PluginInfo* PluginEntryFctn(void)
#endif
{
  return(&httpPlugin);
}
 
